--- Handles loot from the system/data side: mainly calculating GP

-- following tables modified from LibGearPoints-1.2 (EPGP)
local EQUIPSLOT_MULTIPLIER_1 = {
  INVTYPE_HEAD = 1,
  INVTYPE_NECK = 0.5,
  INVTYPE_SHOULDER = 0.75,
  INVTYPE_CHEST = 1,
  INVTYPE_ROBE = 1,
  INVTYPE_WAIST = 0.75,
  INVTYPE_LEGS = 1,
  INVTYPE_FEET = 0.75,
  INVTYPE_WRIST = 0.5,
  INVTYPE_HAND = 0.75,
  INVTYPE_FINGER = 0.5,
  INVTYPE_TRINKET = 1,
  INVTYPE_CLOAK = 0.5,
  INVTYPE_WEAPON = 1.5,
  INVTYPE_SHIELD = 1.5,
  INVTYPE_2HWEAPON = 2,
  INVTYPE_WEAPONMAINHAND = 1,
  INVTYPE_WEAPONOFFHAND = 1,
  INVTYPE_HOLDABLE = 0.5, --off-hand frill
  INVTYPE_RANGED = 2, --bow
  INVTYPE_RANGEDRIGHT = 2, --wand, crossbow or gun
}
local EQUIPSLOT_MULTIPLIER_2 = {
  INVTYPE_WEAPON = 0.5, --tanking = .5, other main hand = 1.5
  INVTYPE_SHIELD = 0.5, --tanking shields = 1.5, shaman = .5
  INVTYPE_RANGEDRIGHT = 1.5, --wands = 1.5, xbow/gun = 2
}

local CUSTOM_ITEM_DATA = {
  -- T16 Normal
  [99685] = { 553, 'INVTYPE_SHOULDER' },
  [99695] = { 553, 'INVTYPE_SHOULDER' },
  [99690] = { 553, 'INVTYPE_SHOULDER' },
  [99684] = { 553, 'INVTYPE_LEGS' },
  [99693] = { 553, 'INVTYPE_LEGS' },
  [99688] = { 553, 'INVTYPE_LEGS' },
  [99683] = { 553, 'INVTYPE_HEAD' },
  [99694] = { 553, 'INVTYPE_HEAD' },
  [99689] = { 553, 'INVTYPE_HEAD' },
  [99682] = { 553, 'INVTYPE_HAND' },
  [99692] = { 553, 'INVTYPE_HAND' },
  [99687] = { 553, 'INVTYPE_HAND' },
  [99696] = { 553, 'INVTYPE_CHEST' },
  [99691] = { 553, 'INVTYPE_CHEST' },
  [99686] = { 553, 'INVTYPE_CHEST' },
  -- T16 Normal Essences
  [105857] = { 553, 'INVTYPE_HEAD' },
  [105859] = { 553, 'INVTYPE_HEAD' },
  [105858] = { 553, 'INVTYPE_HEAD' },
  -- T16 Heroic (Mythic after 6.0)
  [99717] = { 566, 'INVTYPE_SHOULDER' },
  [99719] = { 566, 'INVTYPE_SHOULDER' },
  [99718] = { 566, 'INVTYPE_SHOULDER' },
  [99726] = { 566, 'INVTYPE_LEGS' },
  [99713] = { 566, 'INVTYPE_LEGS' },
  [99712] = { 566, 'INVTYPE_LEGS' },
  [99723] = { 566, 'INVTYPE_HEAD' },
  [99725] = { 566, 'INVTYPE_HEAD' },
  [99724] = { 566, 'INVTYPE_HEAD' },
  [99720] = { 566, 'INVTYPE_HAND' },
  [99722] = { 566, 'INVTYPE_HAND' },
  [99721] = { 566, 'INVTYPE_HAND' },
  [99714] = { 566, 'INVTYPE_CHEST' },
  [99716] = { 566, 'INVTYPE_CHEST' },
  [99715] = { 566, 'INVTYPE_CHEST' },
  -- T17
  [119309] = { 665, "INVTYPE_SHOULDER" },
  [119322] = { 665, "INVTYPE_SHOULDER" },
  [119314] = { 665, "INVTYPE_SHOULDER" },

  [119307] = { 665, "INVTYPE_LEGS" },
  [119320] = { 665, "INVTYPE_LEGS" },
  [119313] = { 665, "INVTYPE_LEGS" },

  [119308] = { 665, "INVTYPE_HEAD" },
  [119321] = { 665, "INVTYPE_HEAD" },
  [119312] = { 665, "INVTYPE_HEAD" },

  [119306] = { 665, "INVTYPE_HAND" },
  [119319] = { 665, "INVTYPE_HAND" },
  [119311] = { 665, "INVTYPE_HAND" },

  [119305] = { 665, "INVTYPE_CHEST" },
  [119318] = { 665, "INVTYPE_CHEST" },
  [119315] = { 665, "INVTYPE_CHEST" },

  [119310] = { 665, "INVTYPE_HEAD" },
  [120277] = { 665, "INVTYPE_HEAD" },
  [119323] = { 665, "INVTYPE_HEAD" },
  [120279] = { 665, "INVTYPE_HEAD" },
  [119316] = { 665, "INVTYPE_HEAD" },
  [120278] = { 665, "INVTYPE_HEAD" }
}

local BONUS_ILVL = {
  [570] = 15, -- heroic tier
  [569] = 30  -- mythic tier
}

local BONUS_GP = {
  -- extra prismatic socket
  [523] = 100,
  [563] = 100,
  [564] = 100,
  [565] = 100,
  [572] = 100
}

local function GetBonuses(link)
  --- returns bonus ilvl, bonus gp
  -- example normal & heroic:
  -- item:119320:0:0:0:0:0:0:0:null:0:0:1:0|h[Leggings of the Iron Protector]|h|r
  -- item:119320:0:0:0:0:0:0:0:null:0:0:1:570|h[Leggings of the Iron Protector]|h|r
  local bonus_ilvl = 0
  local bonus_gp = 0
  local item_string = string.match(link, 'item:[^|]+')
  local components = { strsplit(':', item_string) }
  for i = 14, #components do
    local n = tonumber(components[i])
    if BONUS_ILVL[n] then
      flsys:Info('Found ilvl bonus of ' .. BONUS_ILVL[n] .. ' for ' .. link)
      bonus_ilvl = bonus_ilvl + BONUS_ILVL[n]
    end
    if BONUS_GP[n] then
      flsys:Info('Found gp bonus of ' .. BONUS_GP[n] .. ' for ' .. link)
      bonus_gp = bonus_gp + BONUS_GP[n]
    end
  end
  return bonus_ilvl, bonus_gp
end

function flsys:GetValue(level, itemid, equipLoc, link)
  -- function adapted from LibGearPoints-1.2 (EPGP)

  if flsys.db.mode == 2.2 then
    return 1
  end

  itemid = tonumber(itemid)

  -- tier items
  if CUSTOM_ITEM_DATA[itemid] then
    level, equipLoc = unpack(CUSTOM_ITEM_DATA[itemid])
  end

  -- difficulty (reg/heroic/mythic) and bonus socket
  local bonus_ilvl, bonus_gp = GetBonuses(link)
  level = level + bonus_ilvl

  -- slot multipliers
  local slot_multiplier1 = EQUIPSLOT_MULTIPLIER_1[equipLoc]
  local slot_multiplier2 = EQUIPSLOT_MULTIPLIER_2[equipLoc]
  if not slot_multiplier1 then
    flsys:Warning('No slot multiplier found; returning 0')
    return 0
  end

  -- calculate GP
  local gp_base = 1000 * 2^((level - flsys.db.standard_ilvl)/26)
  local high = math.floor(0.5 + gp_base * slot_multiplier1) + bonus_gp
  local low
  if slot_multiplier2 then
    low = math.floor(0.5 + gp_base * slot_multiplier2) + bonus_gp
  end
  flsys:Info('Returning '..high..' GP value')
  return high, low
end

