--- Utility functions

function flsys:IsEmptyString(str)
  if (#str == 0) then
    return true
  elseif string.find(str, '^%s*$') then
    return true
  else
    return false
  end
end

function flsys:Round(num, dp)
  local mult = 10^(dp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function flsys:TextTooltip(owner, text)
  GameTooltip:SetOwner(owner, "ANCHOR_RIGHT")
  GameTooltip:SetText(text)
  GameTooltip:Show()
end

function flsys:Print(text)
  if text then
    DEFAULT_CHAT_FRAME:AddMessage("|cff6600FFflsys:|r "..tostring(text))
  end
end

local LOWER = {
  a = true,
  an = true,
  the = true,
  of = true,
  on = true,
  ['or'] = true
}

function flsys:Cap(str)
  local capstr = ""
  local first = true
  for word in string.gmatch (str, "[^%s]+") do
    if first or not LOWER[word] then
      word = (word:gsub("^%l", string.upper))
    end
    capstr = capstr..word.." "
  end
  return string.sub(capstr, 1, -2)
end

local LOG_LEVELS = { --TODO implement more readable log system (see below)
  ERROR = 0,
  WARNING = 1,
  INFO = 2,
  DEBUG = 3
}

local LEVEL_TO_STR = {
  [0] = "E",
  [1] = "W",
  [2] = "I",
  [3] = "D"
}

function flsys:_LogToDB(lvl, ...)
  if lvl <= (flsys.db.debuglevel or 3) then
    local level = LEVEL_TO_STR[lvl]
    local log_msg = string.format('%s [%s] %s', date("%m/%d %H:%M:%S"), level, table.concat({...}, " "))
    table.insert(flsys.log, log_msg)
  end
end

function flsys:Debug(...)
  self:_LogToDB(LOG_LEVELS.DEBUG, ...)
end

function flsys:Info(...)
  self:_LogToDB(LOG_LEVELS.INFO, ...)
end

function flsys:Warning(...)
  self:_LogToDB(LOG_LEVELS.WARNING, ...)
end

function flsys:Error(...)
  self:_LogToDB(LOG_LEVELS.ERROR, ...)
end

function flsys:TrimDebug()
  for i = 4001, #flsys.log do
    table.remove(flsys.log, 1)
  end
end

