--- Module to handle loot distribution functions

fld = flsys:NewModule("fld", "AceConsole-3.0", "AceEvent-3.0")

-------------------------------------------------------------------------------
--- VARIABLES
-------------------------------------------------------------------------------
local lootTable = {} -- TODO remove (replaced by loot)
local lootCache = {} -- TODO remove (already in loot)
local lootdone = 0  -- TODO move to loot?
local roster = {} --k = name,v ={rollsdone, row}
local size = 0
local rollingopen = false
local rollsrec = 0
local progressframebuttons = {}
local lootlistbuttons = {[1]="blank"}
local db

local QUEST_ITEMS = {
  [77952] = true, --elementium gem cluster (dragon soul)
  [71141] = true, --eternal ember (firelands)
}
local IGNORE_LIST = {
  ["Sigil of Power"] = true, -- name instead of id to avoid calcing id twice
  ["Sigil of Wisdom"] = true,
  ["Key to the Palace of Lei Shen"] = true,
}

local ROLLORDER = {
  ["m"] = 0, --masterlooter assigned
  ["n"] = 1,
  ["s"] = 2,
  ["o"] = 3,
  ["g"] = 4,
  ["z"] = 5, --sell/shard
  ["p"] = 6,
  ["u"] = 7,
}

local defaults = {
  profile = {
    autoassign = {},
    rarity_threshold = 3,
    sharder = "",
    seller = "",
    boeopts = "np",
    emberopts = "np",
    defaultopts = "nsop"
  }
}

-------------------------------------------------------------------------------
--- LootItem
-------------------------------------------------------------------------------

local LootItem = {}
LootItem.__index = LootItem

LootItem.defaults = {
  needcount = 0,
  _quantity = 1,
  _awarded = 0
}

function LootItem:new(name, link, ilvl, texture, equipLoc, class, subtype, index, table)
  local item
  if table then
    item = table
  else
    item = {}
  end
  item.itemid = fld:LinktoID(link)
  item.name = name
  item.equipLoc = equipLoc
  item.class = class
  item.subtype = subtype
  item.link = link
  item.texture = texture
  item.ilvl = ilvl
  item.index1 = index
  item.rolls = {}
  item.subs = {}
  for k, v in pairs(self.defaults) do
    item[k] = v
  end
  setmetatable(item, LootItem)
  item:set_roll_opts()
  return item
end

function LootItem:set_roll_opts()
  self.rollopts = db.defaultopts
  local assigned = db.autoassign[self.name]
  if assigned and roster[assigned] then
    table.insert(self.rolls, {
      player = db.autoassign[self.name],
      rolltype = "m",
      rollstr = "auto-assigned to "..db.autoassign[self.name],
      roll = 0
    })
    -- TODO call function to do this
    rollsrec = rollsrec+size --for progress purposes, say everyone rolled on it
    flsys:Info("Autoassigned "..self.name.." and incremented rolls received.")
  else
    if fld:IsBoE(self.link) then --boe gear
      flsys:Debug("set_roll_opts - "..self.name.." is boe")
      if db.boeopts then
        self.rollopts = db.boeopts
      end
      self.passtype = "sell"
    elseif self.class =="Weapon" or self.class =="Armor" then --bop gear
      self.passtype = "disenchant"
      flsys:Debug("set_roll_opts - "..self.name.." set to disenchant passtype")
    elseif self.class =="Trade Goods" then
      flsys:Debug("set_roll_opts - "..self.name.." is trade goods")
      if db.emberopts then
        self.rollopts = db.emberopts
      end
      self.passtype = "sell"
    elseif self.class =="Recipe" then --recipe
      flsys:Debug("set_roll_opts - "..self.name.." is a recipe")
      self.rollopts = "gp"
      self.passtype = "sell"
    elseif fld:IsMount(self.link) then --mount
      self.rollopts = "gp"
    elseif QUEST_ITEMS[self.itemid] then --quest item, can't sell or de
      self.rollopts = "nop"
    elseif subtype =="Junk" then --token
      self.passtype ="sell"
    else
      flsys:Warning(self.name.." not categorized; using default opts no passtype")
    end
  end
end

function LootItem:add_copy(i)
  self._quantity = self._quantity + 1
  self['index'..self._quantity] = i
end

function LootItem:clear()
  for k, v in pairs(self) do
    if type(v) == 'table' then
      wipe(self[k])
    elseif self.defaults[k] then
      self[k] = self.defaults[k]
    elseif type(v) ~= 'function' then
      self[k] = nil
    end
  end
end

function LootItem:add_roll(player_name, rolltype, roll, rollstr)
  if (rolltype == 'u' or string.find(self.rollopts, rolltype)) then
    -- TODO clean this up
    if rolltype == 'n' and flsys.db.mode == 1 and not (db.players[player_name].temp and db.players[player_name].gp_total>0) then
      self.needcount = self.needcount + 1
      if db.players[player_name].temp then
        table.insert(self.subs, player_name)
        flsys:Print(player_name.." inserted into sub table at position "..#self.subs)
        flsys:Info(player_name.." inserted into sub table at position "..#self.subs)
      end
    end

    table.insert(self.rolls, {
      player = player_name,
      rolltype = rolltype,
      roll = roll,
      rollstr = rollstr
    })
    flsys:Info('Roll recorded for '..self.name..': '..rollstr)
    return true
  else
    flsys:Warning('Invalid roll type '..rolltype..' for '..self.name)
    return false
  end
end

function LootItem:pre_sort()
  if self.subroll and self.rolls[1].rolltype == "m" and db.players[self.rolls[1].player].gp_total > 0 then
    -- if sub was bumped and since won something, un-bump them
    self.rolls[1].rolltype = "n"
    flsys:Info("PreSortSubs - "..self.rolls[1].player.." has won something; putting back at end of list for "..self.name)
    table.remove(self.subs, self.subroll) -- subroll is index of winning sub
    self.subroll = nil
    self.subannounce = nil
    self.needcount = self.needcount - 1
  end
  if #self.subs > 0 and self.needcount > self._quantity and not self.subroll then
    -- sub roll needed
    self.subroll = math.random(self.needcount)
    flsys:Info("pre_sort_subs - "..self.name.." needers: "..self.needcount..", subs: "..#self.subs.." random roll of "..self.subroll)
    self.subannounce = string.format("%d people rolled need on %s.  Random roll was %d.", self.needcount, self.link, self.subroll)
    --flsys:Print(self.subannounce)
    if self.subs[self.subroll] then
      self:bump_sub()
    end
  end
end

function LootItem:bump_sub()
  local subname = self.subs[self.subroll]
  for i, r in ipairs(self.rolls) do
    if r.player == subname then
      r.rolltype = "m"
      flsys:Print(subname.." won the sub roll for "..self.name.."; bumping to top.")
    end
  end
end

local function compare_rolls(rollA, rollB)
  if rollA.rolltype ~= rollB.rolltype then 
    return ROLLORDER[rollA.rolltype] < ROLLORDER[rollB.rolltype]
  elseif rollA.rolltype == "n" and (flsys:GetPR(rollA.player)~= flsys:GetPR(rollB.player)) then
    return flsys:SortByPR(rollA.player, rollB.player)
  else
    return rollA.roll > rollB.roll
  end
end

function LootItem:sort_rolls()
  --only need to sort if #rolls > #items
  if #self.rolls > self._quantity then
    self:pre_sort()
    table.sort(self.rolls, compare_rolls)
  end
end

function LootItem:process_winner()
  -- handle items that were passed on
  if not self.rolls[1] then
    flsys:Warning('Item has zero rolls including autopass. '..self.name)
    return
  end
  if (self.rolls[1].rolltype == 'p' or self.rolls[1].rolltype == 'u') then
    passtype = self.passtype or ''
    if passtype == 'sell' and db.seller ~= '' then
      table.insert(self.rolls, 1, {
        player = db.seller,
        rolltype = 'z',
        rollstr = ' '..passtype
      })
      flsys:Info('Setting winner for '..self.name..' to seller')
    elseif passtype == 'disenchant' and db.sharder ~= '' then
      table.insert(self.rolls, 1, {
        player = db.sharder,
        rolltype = 'z',
        rollstr = ' '..passtype
      })
      flsys:Info('Setting winner for '..self.name..' to sharder')
    end
  else
    flsys:Info('Winner for '..self.name..' is '..self.rolls[1].player..' with '..self.rolls[1].rolltype)
  end
end

function LootItem:is_autoassigned()
  if self.rolls[1] and self.rolls[1].rolltype == 'm' then
    return true
  else
    return false
  end
end

function LootItem:awarded(overrode)
  self._awarded = self._awarded + 1
  if self._awarded >= self._quantity then
    self.winnerbutton:Hide()
  elseif not overrode then
    -- if there are more copies, remove the person who just got one
    table.remove(item.rolls, 1)
  end
end

function LootItem:amount_to_award()
  return self._quantity - self._awarded
end

function LootItem:current_index()
  return self._awarded + 1
end

--------------------------------------------------------------------------------
--- Loot table
--------------------------------------------------------------------------------

local loot = {
  _mob = '',
  items = {},
  count = 0, -- how many of .items are active (does not include copies)
  cache = {},
  for_rolling = 0, -- how many items are being rolled on
}

function loot:set_mob(name)
  if loot._mob == '' then
    loot._mob = name
  end
  return loot._mob
end

-- returns string|false rollopts if new item, false if copy
function loot:add_item(name, link, ilvl, texture, equipLoc, class, subtype, index)
  if not self:find_copy(link) then
    self.count = self.count + 1
    local create = self.items[self.count] == nil
    self.items[self.count] = LootItem:new(name, link, ilvl, texture, equipLoc, class, subtype, index, self.items[self.count])
    self:cache_item(self.count, link)
    if create then
      fld:create_award_frame_section(self.items[self.count], self.count)
    end
    return self.items[self.count].rollopts
  end
  return false
end

function loot:cache_item(i, link)
  if not self.cache._mob then
    self.cache._mob = self._mob
  end
  local cache_item = self.cache[self.count]
  if cache_item then
    cache_item.link = link
  else
    cache_item = {
      link = link,
      rolls = {}
    }
    self.cache[self.count] = cache_item
  end
end

function loot:in_progress()
  -- true if loot has been announced, and awarding is not yet done
  if self.items[1] and self.items[1].link ~= nil then
    return true
  else
    return false
  end
end

function loot:clear()
  flsys:Info('Clearing loot table')
  self.count = 0
  self.copies = 0
  self._mob = ''
  for i, item in ipairs(self.items) do
    item:clear()
  end
  -- TODO hide frames?
end

function loot:refresh_indices()
  flsys:Info("Refreshing loot table indices")
  local copies, drop_link
  for pos, item in ipairs(self.items) do
    copies = 1
    for i = 1, GetNumLootItems() do
      if GetLootSlotType(i) == 1 then
        drop_link = GetLootSlotLink(i)
        if drop_link == item.link then
          item["index"..copies] = i
          copies = copies + 1
        end
      end
    end
  end
end

function loot:find_copy(link)
  for i, item in ipairs(self.items) do
    if item.link == link then
      flsys:Info("Looting", "Found a copy! "..link)
      item:add_copy(i)
      return i
    end
  end
  return false
end

loot.rolltype_tostring = {
  ["n"] = "%s (need)", --player
  ["p"] = "%s (pass)", --player
  ["s"] = "%s (sidegrade, rolled %d)", --player, roll
  ["g"] = "%s (greed, rolled %d)", --player, roll
  ["o"] = "%s (offspec, rolled %d)", --player, roll
  ["u"] = "%s (autopass)", --player
}
function loot:add_roll(player_name, itemid, rolltype, roll)
  local match
  for i, item in pairs(self.items) do
    if item.itemid == itemid then
      match = item
      break
    end
  end
  if match then
    local rollstr = string.format(self.rolltype_tostring[rolltype], player_name, roll)
    return match:add_roll(player_name, rolltype, roll, rollstr)
  else
    flsys:Error('No matching item found for roll from '..player_name..' '..itemid)
    return false
  end
end

function loot:clear_cache()
  self.cache._mob = ''
  for i, cache_item in ipairs(self.cache) do
    cache_item.link = nil
    wipe(cache_item.rolls)
  end
end

function loot:cache_rolls(i)
  local cache_item = self.cache[i]
  if cache_item then
    for j, roll in ipairs(self.items[i].rolls) do
      cache_item.rolls[i] = roll.rollstr
    end
  else
    flsys:Error('No cache item corresponding to loot item. Index: '..i)
  end
end

function loot:cache_to_string(i)
  local printstr
  if i then
    if self.cache[i].link then
      if self.cache[i].rolls[1] then
        printstr = "Rolls for "..self.cache[i].link..":\n"
        for i, s in ipairs(self.cache[i].rolls) do
          printstr = printstr..i..". "..self.cache[i].rolls[i].."\n"
        end
      else
        printstr = "No rolls stored for "..self.cache[i].link
      end
    else
      printstr = "There is no cached item under that number"
    end
  elseif self.cache[1].link then
    printstr = self.cache._mob
    for i, cache_item in ipairs(self.cache) do
      if cache_item.link then
        printstr = printstr..i..". "..lootCache[i].link.."\n"
      end
    end
  end 
  return printstr
end

function loot:sort_rolls()
  flsys:Info('Sorting all rolls')
  for i, item in ipairs(self.items) do
    if item.link and item:amount_to_award() > 0 then
      item:sort_rolls()
      item:process_winner()
      self:cache_rolls(i)
    end
  end
  fld:UpdateAwardFrame()
end

function loot:get_addon_message()
  local msg = 'loot:'
  for i, item in ipairs(self.items) do
    if not item.link then
      break
    end
    if item.rollopts and not item:is_autoassigned() then
      msg = msg..item.itemid.."="..item.rollopts..";"
      self.for_rolling = self.for_rolling + 1
    end
  end
  if msg == 'loot' then
    return false
  else
    return msg
  end
end

--[[

local Raider = {}
Raider.__index = Raider

function Raider:New(name)
  return {
    rollsdone = 0
  }
end

local roster = {}

-- reset
-- add player
      
  lootTable -> loot
  lootcount
  rollingopen rolling_open
  count (in various locals)
]]

-------------------------------------------------------------------------------
---ENABLE
-------------------------------------------------------------------------------
function fld:OnEnable()
  fld.db = LibStub("AceDB-3.0"):New("FlamingLootSysDB", defaults)
  db = fld.db.profile
  RegisterAddonMessagePrefix("flamingloot")
  self:RegisterEvent("CHAT_MSG_ADDON")
  self:RegisterChatCommand("loot", "AccessCache")
  flsys:Info('fld enabled *****')
  if TESTING then
    fld.loot = loot
    fld.LootItem = LootItem
  end
end

-------------------------------------------------------------------------------
--- START RAID
-------------------------------------------------------------------------------

function fld:StartRaid() --called by flsys
  self:RegisterEvent("LOOT_OPENED")
  flsys:Debug("fld:StartRaid(), registering for LOOT_OPENED event")

  fld:CreateRoster()

  if fld.db.profile.promptz and not roster[db.sharder] then
    fld:CheckSharder()
  end

  self:RegisterEvent("CHAT_MSG_WHISPER")
  if fld.db.profile.whispopts then
    flsys:Print("Announcing roll options for people without the addon is enabled.  Open the options menu to disable.")
  else
    flsys:Print("Announcing roll options for people without the addon is disabled.  Open the options menu to enable.")
  end
end

function fld:CreateRoster()
  wipe(roster)
  size = GetNumGroupMembers()
  for i = 1, GetNumGroupMembers() do
    local name = select(1, GetRaidRosterInfo(i))
    roster[name] = {["rollsdone"] = 0}
  end
  fld:InitializeFrames()
end

local function createProgressRow(texttable)
  --- Creates a row in the progress frame, uses given text
  local row = CreateFrame("Button", nil, FLProgressFrame)
  row:SetSize(100,16)
  row.cells = {}
  for i, w in ipairs({80, 20}) do
    local c = row:CreateFontString(nil, "ARTWORK", "GameFontWhite")
    c:SetJustifyH("LEFT")
    c:SetWidth(w)
    if i == 1 then
      c:SetPoint("LEFT", row, "LEFT", 0, 0)
    else
      c:SetPoint("LEFT", row.cells[1], "RIGHT", 0, 0)
    end
    table.insert(row.cells, c)
  end
  return row
end

function fld:InitializeFrames()
  --progress frame
  FLProgressFrame:SetScale(.8)
  for i = 1, size do
    if not progressframebuttons[i] then
      local row = createProgressRow()
      row:SetPoint("TOPLEFT", FLProgressFrame, 10, -15-16*i)
      table.insert(progressframebuttons, row)
    end
  end
  FLProgressFrame:SetSize(120, 70 + size * 17)
  local i = 0
  for name, _ in pairs(roster) do
    --assigns rows, so they can be reused when we re-make the roster
    i = i + 1
    progressframebuttons[i].cells[1]:SetText(name)
    roster[name].row = i
  end

  --awarding frame
  FLAwardFrame:SetScale(.8)
  FLAwardFrame.TitleText:SetText("Award loot")
  FLAwardFrame:SetScript("OnHide", function(self)
      flsys:Debug("FLAwardFrame OnHide")
      fld:HideRollersList()
    end)
  FLAwardFrame.CloseButton:SetScript("OnClick", function() fld:EndAwardingClicked() end)

  --rollers list
  FLRollersList:SetHitRectInsets(-10, 10, 10, -10)
end

-------------------------------------------------------------------------------
--- LOOTING
-------------------------------------------------------------------------------

function fld:LOOT_OPENED()
  --- event handler for loot window:
  -- checks the loot table for anything above rarity threshold
  -- if re-opening a window, refreshes the loot indices
  -- otherwise, calls StaticPopup STARTROLLING
  local mob = (UnitName("target") or "Loot chest")
  flsys:Info("LOOT_OPENED handler called, target is "..mob)
  if loot:in_progress() then -- already opened
    flsys:Print("The loot table is not empty - awarding not ended.  Not announcing loot.")
    flsys:Info("Loot table not empty (awarding not ended). Not announcing loot.")
    loot:refresh_indices()
  else
    for i = 1, GetNumLootItems() do
      -- checks
      if GetLootSlotType(i) == 1 then -- TODO constant
        local itemlink = GetLootSlotLink(i)
        local rarity = select(3, GetItemInfo(itemlink))
        if rarity >= db.rarity_threshold and not IGNORE_LIST[itemname] then
          -- passed checks, check announce
          self:RegisterEvent("LOOT_CLOSED")
          flsys:Info("Loot passed checks. Registered for LOOT_CLOSED. Prompting to roll.")
          StaticPopup_Show("STARTROLLING") -- TODO break?
          break
        else
          flsys:Info("Loot didn't pass rarity check")
        end
      end
    end
  end
end

function fld:get_loot_info()
  --- iterates through loot, populates loot, announces loot in chat
  local announcemsg = loot:set_mob((UnitName("target") or "Loot chest"))
  local announcemsg = announcemsg..": "
  local announcepos = 1
  loot:clear_cache()
  for i = 1, GetNumLootItems() do
    if GetLootSlotType(i) == 1 then
      local itemlink = GetLootSlotLink(i)
      local itemname, _, rarity, ilvl, _, class, subtype, _, equipLoc, texture, _ = GetItemInfo(itemlink)
      subtype = subtype or "no subtype"
      if rarity >= db.rarity_threshold and not IGNORE_LIST[itemname] then
        local rollopts = loot:add_item(itemname, itemlink, ilvl, texture, equipLoc, class, subtype, i)
        announcemsg = announcemsg..itemlink.." "
        if fld.db.profile.whispopts and rollopts then
          announcemsg = announcemsg.."("..rollopts..") "
        end
        -- sends announce in raid chat
        SendChatMessage(string.sub(announcemsg, announcepos, -1), "RAID", nil)
        flsys:Info("Announced loot in chat: "..string.sub(announcemsg, announcepos, -1))
        announcepos = string.len(announcemsg)
      end
    end
  end
  if loot.count > 0 then
    fld:StartRolling()
  else
    flsys:Warning("lootcount 0 at end of getlootinfo")
  end
end


function fld:StartRolling()
  local msg = loot:get_addon_message()
  if msg then
    rollingopen = true
    SendAddonMessage("flamingloot", msg, "RAID")
    flsys:Info("StartRolling *** Sent addon message to raid: "..msg)
    fld:ShowProgressFrame(loot.for_rolling)
  else
    --must be only autoloot
    flsys:Warning("StartRolling, nothing to roll on but loot.count is "..loot.count)
    fld:UpdateAwardFrame()
    FLAwardFrame:Show()
  end
end


function fld:LOOT_CLOSED()
  if rollingopen or FLAwardFrame:IsShown() then
    --rolls/looting in progress, warn that window must be open to award
    if rollingopen then
      flsys:Warning("Loot window closed while rolling is open. Showing warning")
    else
      flsys:Warning("Loot window closed while awarding frame is shown. Showing warning")
    end
    StaticPopup_Show("LOOT_WINDOW_AWARD")
  else -- passed rarity check, haven't finished GetLootInfo yet
    -- clear loot in case it was partway through doing GetLootInfo
    loot:clear()
    -- show warning to reopen in order to proceed
    flsys:Warning("Loot window closed - had not accepted rolling prompt yet. Showing warning")
    StaticPopup_Show("LOOT_WINDOW_ANNOUNCE")
  end
end

-------------------------------------------------------------------------------
--- ITEM INFO
-------------------------------------------------------------------------------

function fld:LinktoID(itemlink)
  local itemid = string.match(itemlink, "item:(%d+)")
  return tonumber(itemid)
end

function fld:IsBoE(link) --adapted from LibItemUtils-1.0 (EPGP)
  FLScanningTooltip:SetHyperlink(link)
  local text = FLScanningTooltipTextLeft2:GetText()
  FLScanningTooltip:ClearLines()
  if text then
    if text == ITEM_BIND_ON_EQUIP then
      flsys:Debug(link.." is BoE")
      return true
    else
      return false
    end
  end
end

function fld:IsMount(link)
  FLScanningTooltip:SetHyperlink(link)
  local three = FLScanningTooltipTextLeft3:GetText()
  local four = FLScanningTooltipTextLeft4:GetText()
  if (three and three =="Mount") or (four and four =="Mount") then
    flsys:Debug(link.." is mount")
    return true
  else
    return false
  end
end

-------------------------------------------------------------------------------
--- ROLLING
-------------------------------------------------------------------------------
function fld:ShowProgressFrame(count)
  FLProgressFrameTitle:SetText(count.." items")
  flsys:Info("Showing progress frame for "..count.." items")
  for player,v in pairs(roster) do
    fld:UpdateProgressFrame(player)
  end
  FLProgressFrame:Show()
end

function fld:UpdateProgressFrame(player)
  if roster[player].rollsdone == loot.count then
    progressframebuttons[roster[player].row].cells[2]:SetText("|cff98e0ff"..roster[player].rollsdone)
    progressframebuttons[roster[player].row].cells[1]:SetText("|cff98e0ff"..player)
  else
    progressframebuttons[roster[player].row].cells[1]:SetText(player)
    progressframebuttons[roster[player].row].cells[2]:SetText(roster[player].rollsdone)
  end
end

function fld:CHAT_MSG_ADDON(event, prefix, message, channel, sender)
  if (prefix == "flamingloot") then
    local colon = strfind(message,":")
    local command = strsub(message,1, colon - 1)
    local param = strsub(message, colon + 1)
    if command == "roll" then
      sender = string.match(sender, "([^-]+)-")
      fld:roll_received(sender, param)
    end
  end
end

function fld:CHAT_MSG_WHISPER(bl, message, sender)
  if not roster[sender] then
    return
  end
  if message =="addon" then
    flsys:Info("addon whisper from "..sender)
    SendChatMessage("Download flaming loot at curse.com/addons/wow/flaming or roll via whisper. Loot is announced in raid chat along with roll options. N = need, s = sidegrade (mainspec, same ilvl or lower), o = offspec, g = greed. Rolling opens as soon as loot has been linked.", "WHISPER", nil, sender)
    SendChatMessage("To roll, send me '[itemlink] n' where n is your selection (whole word or just the first letter).  You'll get a reply stating whether your roll was received.  If you are a sub, send 'sub' for an explanation of loot rules.", "WHISPER", nil, sender)
  elseif message =="sub" then
    flsys:Info("sub whisper from "..sender)
    SendChatMessage("Subs are welcome to roll on any item. Sidegrade, OS & greed are random rolls. For need, you're assigned a number & the addon does a random roll between 1 and the number of needs on the item. If the roll matches your number, you win the item; if not", "WHISPER", nil, sender)
    SendChatMessage("it goes to the person with the highest priority in our system. This gives you the same odds as an open roll. Once you have won a mainspec item you'll only be eligible for another one if no one else rolls need. Please whisper me if you have any questions!", "WHISPER", nil, sender)
  else
    local itemid, selec = string.match(message, "item:(%d+).+|h|r%s*(%a)")
    if itemid and selec then
      if rollingopen then
        local r = math.random(100)
        flsys:Info("Roll received via whisper")
        fld:roll_received(sender, itemid.."="..selec..";"..r, true)
      else
        flsys:Info("Roll received via whisper but rolling is closed")
        SendChatMessage("Rolling is not open.", "WHISPER", nil, sender)
      end
    end
  end
end

function fld:roll_received(sender, message, whisp)
  --message format: "itemid=rolltype;roll"
  local itemid, rolltype, roll = string.match(message, "(%d+)=(%a+);*(%d*)")
  roll = tonumber(roll) or 0
  itemid = tonumber(itemid)

  local valid = loot:add_roll(sender, itemid, rolltype, roll)

  if valid then
    if whisp then
      SendChatMessage("Roll received"..string.format(stringconverter[rolltype], "", roll), "WHISPER", nil, sender)
      flsys:Info("Sent whisper confirmation of roll received")
    end
    roster[sender].rollsdone = roster[sender].rollsdone + 1 --for tracking progress
    fld:UpdateProgressFrame(sender)
    rollsrec = rollsrec + 1

    if rollsrec == size * loot.count then
      flsys:Info("All rolls received *** ending looting automatically.")
      fld:close_rolling()
    end

  elseif whisp then
    SendChatMessage("Invalid roll! Follow the format '[itemlink] n' where n is your selection.  Make sure you linked the correct item and selected one of its available roll options.", "WHISPER", nil, sender)
    flsys:Info("Warned of invalid whisper roll.")
  else
    flsys:Error('Invalid roll via addon. From '..sender..', '..message)
  end
end

function fld:close_rolling()
  flsys:Info("Rolling closed ***")
  rollingopen = false
  SendAddonMessage("flamingloot", "end:", "RAID")
  SendChatMessage("Rolling closed.", "RAID", nil)
  FLProgressFrame:Hide()
  flsys:Debug("Progress frame hidden")
  loot:sort_rolls()
  FLAwardFrame:Show()
end


-------------------------------------------------------------------------------
--- AWARD
-------------------------------------------------------------------------------

function fld:create_award_frame_section(item, position)
  item.itembutton = CreateFrame("Button", nil, FLAwardFrame)
  item.itembutton:SetSize(30,30)
  item.itembutton:SetPoint("TOPLEFT", FLAwardFrame, "TOPLEFT", 10, -30-(position-1)*40)
  item.itembutton:SetNormalTexture(item.texture)
  item.itembutton:SetScript("OnEnter", function(self, motion) FlamingLoot:ShowTooltip(FLAwardFrame, item.link, "ANCHOR_LEFT") end)
  item.itembutton:SetScript("OnLeave", function(self, motion) GameTooltip:Hide() end)
  item.itembutton:SetText(" ")
  local fontstr = item.itembutton:GetFontString()
  fontstr:SetAllPoints(item.itembutton)
  fontstr:SetFont("Fonts\\ARIALN.TTF", 10, "THICKOUTLINE")

  item.winnerbutton = CreateFrame("Button", nil, FLAwardFrame)
  item.winnerbutton:SetSize(157, 30)
  item.winnerbutton:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight", "ADD")
  item.winnerbutton:SetPoint("TOPLEFT", FLAwardFrame, "TOPLEFT", 43, -30-(position-1)*40)
  item.winnerbutton:SetScript("OnEnter", function(self, motion) fld:ShowRollersList(position) end)
  item.winnerbutton:SetNormalFontObject("GameFontWhiteSmall")
  item.winnerbutton:SetText("blank")
  fontstr = item.winnerbutton:GetFontString()
  fontstr:SetAllPoints(item.winnerbutton)
end

function fld:UpdateAwardFrame()
  -- Once loot table is populated. Sets textures, scripts and size of award frame
  flsys:Info("Updating award frame for "..loot.count.." items")
  for i, item in ipairs(loot.items) do
    local count = item:amount_to_award()
    if count > 0 then
      item.itembutton:SetNormalTexture(item.texture)
      item.itembutton:SetScript("OnEnter", function(self, motion) FlamingLoot:ShowTooltip(FLAwardFrame, item.link, "ANCHOR_LEFT") end)
      if (item.rolls[1]) then
        item.winnerbutton:SetText(item.rolls[1].rollstr)
        item.winnerbutton:SetScript("OnClick", function() fld:AwardPressed(item, item.rolls[1].player, item.rolls[1].rolltype) end)
      else
        item.winnerbutton:SetText('No rolls received')
        item.winnerbutton:SetScript("OnClick", function() end)
      end
      item.winnerbutton:Show()
      item.itembutton:Show()
    end
    if count > 1 then
      item.itembutton:SetText(count)
      flsys:Info(count..' copies of '..item.name)
    end
  end
  FLAwardFrame:SetHeight(30+(loot.count)*40)
end

function fld:CreateRollListButton(pos)
  local b = CreateFrame("button", "RollListButton"..pos, FLRollersList)
  pos = pos - 2 --since winning roll is already shown in main frame
  pos = -pos*16-3
  b:SetSize(190, 20)
  b:SetPoint("TOP", FLRollersList, "TOP", 3, pos)
  b:SetNormalFontObject("GameFontHighlightSmallLeft")
  b:SetHighlightFontObject("GameFontNormalSmallLeft")
  b:SetHighlightTexture("Interface\\Buttons\\UI-Listbox-Highlight2")
  table.insert(lootlistbuttons, b)
end

function fld:ShowRollersList(pos)
  fld:HideRollersList()
  local item = loot.items[pos]
  if item.rolls[2] then
    local count = 0
    for i, rolltable in ipairs(item.rolls) do
      if i > 1 then
         if not lootlistbuttons[i] then
          fld:CreateRollListButton(i)
         end
         lootlistbuttons[i]:SetText(i..". "..rolltable.rollstr)
         lootlistbuttons[i]:SetScript("OnClick", function(self)
             flsys:Debug("Award pressed from rollers list (not main button)")
             fld:AwardPressed(pos, rolltable.player, rolltable.rolltype)
          end)
         lootlistbuttons[i]:Show()
         count = count+1
      end
    end
    FLRollersList:SetHeight(count*16+12)
    FLRollersList:SetPoint("TOPLEFT", item.winnerbutton, "TOPRIGHT", 10, 3)
    FLRollersList:Show()
  end
end

function fld:HideRollersList()
  for i, button in ipairs(lootlistbuttons) do
    if i > 1 then
      button:Hide()
    end
  end
  FLRollersList:Hide()
end

function fld:AwardPressed(item, player, rolltype)
  fld:HideRollersList()
  local value, val2
  if rolltype == "n" then
    value, val2 = flsys:GetValue(item.ilvl, item.itemid, item.equipLoc, item.link)
    if val2 then
      value = value.." or "..val2
    end
  else
    value = 0
  end
  flsys:Info("Award pressed. Showing confirmation dialog with player: "..player..", item: "..item.name)
  local dialog = StaticPopup_Show("CONFIRM_MASTERLOOT",
                                  item.link, player,
                                  {value = value, rolltype = rolltype})
  if dialog then
    dialog.item = item
    dialog.player = player
  end
end

function fld:AwardLoot(item, player, value, rolltype)
  local lootindex  = item:current_index()
  local playerindex -- different per player per piece of loot
  for p = 1, GetNumGroupMembers() do
    if GetMasterLootCandidate(lootindex, p) == player then
      playerindex = p
    end
  end

  -- error cases
  if not GetLootSlotLink(lootindex) then
    flsys:Warning("Couldn't find lootindex for item at position "..pos..". Showing popup")
    StaticPopup_Show("WHEREISLOOT", item.link)
  elseif not playerindex then -- player is not candidate
    flsys:Warning(player..' not a candidate for '..item.name)
    StaticPopup_Show('NOT_ML_CANDIDATE')
    
  -- success - award loot
  else
    local announcemsg = item.link.." - "
    local overrode = false
    if player == item.rolls[1].player then
      -- include rolls details if awarding to top sorted player
      announcemsg = announcemsg..item.rolls[1].rollstr
    else
      -- ml overrode sort
      announcemsg = announcemsg..player
      overrode = true
    end

    GiveMasterLoot(lootindex, playerindex)
    SendChatMessage(announcemsg, "RAID", nil)

    flsys:Info("Awarded "..item.name.." to "..player.." for "..value.." GP ***")
    item:awarded(overrode)
    flsys:OnLootAwarded(player, item.link, item.itemid, value)
    lootdone = lootdone + 1
    FLScanningTooltip:ClearLines() -- TODO timing?

    if lootdone == loot.count then
      flsys:Debug('All items have been awarded. Calling EndAwarding() automatically')
      fld:EndAwarding()
    else
      loot:sort_rolls()
    end
  end
end

function fld:EndAwardingClicked()
  flsys:Info("Close button clicked to end awarding early. Showing confirmation")
  StaticPopup_Show("fldL_CONFIRM_ENDAWARD")
end

function fld:EndAwarding()
  FLScanningTooltip:ClearLines() -- TODO timing?
  flsys:Info("EndAwarding() ***")
  self:UnregisterEvent("LOOT_CLOSED")
  loot:clear()
  for name, nt in pairs(roster) do
    nt.rollsdone = 0
  end
  rollsrec = 0
  FLProgressFrame:Hide()
  lootdone = 0
  FLAwardFrame:Hide()
end

-------------------------------------------------------------------------------
--- SLASH COMMANDS
-------------------------------------------------------------------------------
function fld:AccessCache(input)
  --- Shows recent drops & rolls
  -- Without a number input, shows list of items from last boss/mob/chest
  -- With number, shows all the rolls on that item 
  local num = string.match(input, "(%d)")
  if num then
    num = tonumber(num)
  else
    num = nil
  end
  local printstr = loot:cache_to_string(num)
  if not printstr then
    printstr = "No loot cached."
  elseif printstr and not num then
    printstr = printstr.."\n Type /loot [number] to view rolls for that item."
  end
  flsys:Print(printstr)
end

-------------------------------------------------------------------------------
--- CONFIRMATION
-------------------------------------------------------------------------------
function fld:CheckSeller()
  StaticPopup_Show("SET_SELLER")
end

function fld:CheckSharder()
  StaticPopup_Show("SET_SHARDER")
end

StaticPopupDialogs["CONFIRM_MASTERLOOT"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Give %s to %s?",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  enterClicksFirstButton = true,
  hideOnEscape = 1,
  hasEditBox = 1,
  maxLetters = 15,
  OnShow = function(self, data)
    self.editBox:SetText(tostring(data.value))
    self.editBox:HighlightText()
  end,
  OnAccept = function(self, data)
    local value = self.editBox:GetText()
    if not string.match(value, "^%d+$") then
      flsys:Warning('Accepted award prompt with invalid value: '..value)
      StaticPopup_Show("ENTERNUMBER")
    else
      value = tonumber(value)
      fld:AwardLoot(self.item, self.player, value, data.rolltype)
    end
  end
}

StaticPopupDialogs["ENTERNUMBER"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Please enter a number.",
  button1 = ACCEPT,
  enterClicksFirstButton = true,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
}

StaticPopupDialogs["fldL_CONFIRM_ENDAWARD"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "End awarding early?",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  enterClicksFirstButton = true,
  OnAccept = function(self)
    flsys:Info('Confirmed end awarding early ***')
    fld:EndAwarding()
  end
}

StaticPopupDialogs["SET_SELLER"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Enter a name if you would like to assign someone in the raid to sell unneeded BoE items.",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  hasEditBox = 1,
  maxLetters = 15,
  enterClicksFirstButton = true,
  OnAccept = function(self)
    local text = flsys:Cap(self.editBox:GetText())
    if roster[text] then
      fld.db.profile.seller = text
      flsys:Print(text.." set as the designated seller.")
    else
      flsys:Print("Name did not match a player in the raid. You can set a new name in the options menu.")
    end
  end
}

StaticPopupDialogs["SET_SHARDER"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Enter a name if you would like to set a designated disenchanter for the raid.",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  enterClicksFirstButton = true,
  whileDead = 1,
  hideOnEscape = 1,
  hasEditBox = 1,
  maxLetters = 15,
  OnAccept = function(self)
    local text = flsys:Cap(self.editBox:GetText())
    if roster[text] then
      fld.db.profile.sharder = text
      flsys:Print(text.." set to disenchant unwanted items.")
    else
      flsys:Print("Name did not match a player in the raid. You can set a new name in the options menu.")
    end
    if not roster[fld.db.profile.seller] then
      fld:CheckSeller()
    end
  end,
  OnCancel = function(self)
    if not roster[fld.db.profile.seller] then
      fld:CheckSeller()
    end
  end
}

StaticPopupDialogs["NOT_ML_CANDIDATE"] = {
  text = "%s is not a master loot candidate for %s",
  button1 = ACCEPT,
  button2 = nil,
  sound = "igQuestLogAbandonQuest",
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  cancels = "STARTROLLING",
  enterClicksFirstButton = true,
  OnAccept = function() end
}

StaticPopupDialogs["LOOT_WINDOW_ANNOUNCE"] = {
  text = "Loot window closed before items could be scanned. Loot the mob again and make sure the window is still open when you accept the \"begin looting\" dialog.",
  button1 = ACCEPT,
  button2 = nil,
  sound = "igQuestLogAbandonQuest",
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  cancels = "STARTROLLING",
  enterClicksFirstButton = true,
  OnAccept = function() end
}

StaticPopupDialogs["LOOT_WINDOW_AWARD"] = {
  text = "Loot window closed before awarding has finished.\nThe window will need to be open in order for you to hand out loot.",
  button1 = ACCEPT,
  button2 = nil,
  sound = "igQuestLogAbandonQuest",
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  enterClicksFirstButton = true,
  OnAccept = function() end
}

StaticPopupDialogs["WHEREISLOOT"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "%s not found.  Make sure you have the loot window opened.",
  button1 = ACCEPT,
  button2 = nil,
  sound = "igQuestLogAbandonQuest",
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  enterClicksFirstButton = true,
  OnAccept = function() end
}

StaticPopupDialogs["STARTROLLING"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Begin looting?",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  enterClicksFirstButton = true,
  OnAccept = function(self)
    flsys:Info("STARTROLLING prompt accepted.")
    fld:get_loot_info()
  end,
  OnCancel = function(self)
    flsys:Info('Declined STARTROLLING')
    flsys:UnregisterEvent('LOOT_CLOSED')
  end
}

-------------------------------------------------------------------------------
---- TESTING
-------------------------------------------------------------------------------

-- These functions are accessed directly with a /script command in-game
-- Do fld:PreCacheLoot() first to get items into local cache
-- Reload UI afterward to reset WoW mocked functions

local testitems = {
  {["itemid"] = 119320},  -- t17 legs protector
  {["itemid"] = 113608},  -- 655 mail head
  {["itemid"] = 113648},  -- 655 plate legs
  {["itemid"] = 113859},  -- 655 int trinket
  {["itemid"] = 113666},  -- 665 str sheild
}
fld.testitems = testitems

local cached = false
function fld:PreCacheLoot()
  flsys:Print('Prepping by caching loot items...')
  for k, v in pairs(testitems) do
    print(GetItemInfo(v.itemid))
  end
  flsys:Print('Loot items should now be in your cache. Type /fs test again to start simulated loot rolling.')
  cached = true
end

function fld:FakeLootOpened()
  flsys:Print('Simulating loot rolls. IMPORTANT: After the test, reload your UI. (/reload)')
  local announcemsg = "Test loot: "

  loot:clear()
  loot:clear_cache()

  -- mock
  local origGetNum = GetNumLootItems
  GetNumLootItems = function()
    return #testitems
  end
  local origGetType = GetLootSlotType
  GetLootSlotType = function(i)
    return 1
  end
  local origGetLink = GetLootSlotLink
  function GetLootSlotLink(i)
    return select(2, GetItemInfo(testitems[i].itemid))
  end

  local origGetCandidate = GetMasterLootCandidate
  GetMasterLootCandidate = function(lootindex, p)
    return GetRaidRosterInfo(p);
  end
  local origGiveMasterLoot = GiveMasterLoot
  GiveMasterLoot = function(lootindex, playerindex)
    flsys:Print('Mock awarding loot ' .. lootindex .. ' to player ' .. playerindex);
  end
  fld:LOOT_OPENED()
end

function fld:test()
  if not flsys.db.inprogress then
    flsys:Print('You must start a raid first. Once you\'re the ML in a raid group (number of other people doesn\'t matter), type /fss and hit Start')
  else
    if cached then
      fld:FakeLootOpened()
    else
      fld:PreCacheLoot()
    end
  end
end
