-- Wrapper for all the UI components
flsys.ui = {}
local ui = flsys.ui
local db --TODO should not use db

function ui:Initialize()
  db = flsys.db
  self:InitializeStandings()
  self:InitializeLog()
  self:InitializeRoster()
  self:InitializeOptions()
end

-------------------------------------------------------------------------------
--- STANDINGS
-------------------------------------------------------------------------------
--- Handles the main window containing player standings

local rowbuttons = {}
local maindisplay = 1
local rosterpage = 0

-- initialize -> checkrows, mainbuttons
-- update -> update assumes there are correct # of rows, sets text ->
--   rosterpage
-- sys:refreshsortlist -> checkrows, updatepr
-- startraid -> mainbuttons, createroster
-- endraid -> refreshsortlist
-- 
function ui:InitializeStandings()
  -- frame setup that wasn't done in the xml. only needs to be called once.
  FLMain:SetScale(.8)
  FLMainPrev:GetNormalTexture():SetRotation(-math.pi/2)
  FLMainPrev:SetScript("OnClick", function() ui:RosterPage(-1) end)
  FLMainNext:GetNormalTexture():SetRotation(math.pi/2)
  FLMainNext:SetScript("OnClick", function() ui:RosterPage(1) end)
  local x = 0
  local texts = {" ", "EP", "GP", "PR"}
  local widths = {77, 42, 42, 42}
  heading= CreateFrame("Button", "Heading" , FLMain)
  heading:SetSize(200,14)
  for i, w in ipairs(widths) do
    local c = FLMain:CreateFontString(nil, "ARTWORK", "GameFontNormal")
    c:SetParent(Heading)
    c:SetWidth(w)
    c:SetJustifyH("LEFT")
    c:SetPoint("TOPLEFT", FLMain, "TOPLEFT", 15 + x, -42)
    c:SetText(texts[i])
    x = x + w
  end
  heading:Show()
  FLMainTitleText:SetText("         Standings") -- TODO move to xml
  SetPortraitToTexture(FLMainPortrait, "Interface\\Icons\\Achievement_Arena_5v5_6") --TODO move to xml
  FLMainButton3:SetScript("OnClick", function() ui:ShowLogFrame() end)
  -- state/data-dependent display
  ui:MainButtons()
  ui:RenderStandings()
end


-- CALL ON init, raid start, raid end
function ui:MainButtons()
  -- sets function of main buttons, depending on whether raid is in progress
  if not db.inprogress then
    FLMainButton1:SetText("Start")
    FLMainButton1:SetScript("OnClick", function() StaticPopup_Show("STARTRAID") end)
    FLMainButton2:SetText("Export")
    FLMainButton2:SetScript("OnClick", function() ui:ShowExportFrame("standings") end)
  else
    FLMainButton1:SetText("End")
    FLMainButton1:SetScript("OnClick", function() StaticPopup_Show("ENDRAID") end)
    FLMainButton2:SetText("Mass EP")
    FLMainButton2:SetScript("OnClick", function() ui:ShowMassEP() end)
  end
end

function ui:ShowMassEP()
  if db.inprogress then
    flsys:Info('Showing mass ep prompt')
    StaticPopup_Show('MASS_EP')
  else
    flsys:Warning('Tried to mass ep outside of raid')
    flsys:Print('You must have a raid in progress to do that.')
  end
end

function ui:HideMain()
  FLMain:Hide()
end

local function playerDetails(name)
  local status, reason = flsys:GetPlayerDetails(name)
  return status..'\n\n'..reason
end

-- call on init, pr updated, player added/removed, raid update, reset data
-- (any change to roster or pr)
function ui:RenderStandings()
  --- update the player rows - text and click handling
  ui:CheckRows()
  for i, name in ipairs(db.sortlist) do
    local b = rowbuttons[i]
    if db.players[name].inraid then
      b.cells[1]:SetText("|cffCCFFFF"..name)
      b.cells[2]:SetText("|cffCCFFFF"..db.players[name].ep_total)
      b.cells[3]:SetText("|cffCCFFFF"..db.players[name].gp_total)
      b.cells[4]:SetText("|cffCCFFFF"..db.players[name].current_pr)
      if db.players[name].temp then
        b.cells[2]:SetText("|cffCCFFFF-")
        b.cells[4]:SetText("|cffCCFFFF-")
      end
    else
      b.cells[1]:SetText(name)
      b.cells[2]:SetText(db.players[name].ep_total)
      b.cells[3]:SetText(db.players[name].gp_total)
      b.cells[4]:SetText(db.players[name].current_pr)
    end
    b:SetScript('OnEnter', function(self)
      GameTooltip:SetOwner(self, 'ANCHOR_RIGHT')
      GameTooltip:SetText(name..' - '..playerDetails(name))
      GameTooltip:SetWidth(500)
      GameTooltip:Show()
    end)
    b:SetScript('OnLeave', function() GameTooltip:Hide() end)

  end
  ui:RosterPage(0)
end


local function CreateTableRow(texttable)
  local widths = {77,42,42,42}
  local row = CreateFrame("Button", nil, FLMain)
  row:SetSize(200,14)
  row.cells = {}
  for i,w in ipairs(widths) do
    local c = row:CreateFontString(nil, "ARTWORK", "ChatFontSmall")
    c:SetJustifyH("LEFT")
    c:SetWidth(w)
    if #row.cells == 0 then
      c:SetPoint("LEFT", row, "LEFT", 0, 0)
    else
      c:SetPoint("LEFT", row.cells[#row.cells], "RIGHT", 0, 0)
    end
    c:SetText(texttable[i])
    table.insert(row.cells, c)
  end
  return row
end


-- call when sortlist size changes
function ui:CheckRows()
  ---creates & places row per sortlist if there isn't one already
  for i=1, #db.sortlist do
    if not rowbuttons[i] then
      local row = CreateTableRow({"-", "-", "-", "-"}, widths, FLMain)
      if i%13 ~= 0 then
        row:SetPoint("TOPLEFT", FLMain, "TOPLEFT", 18, -53-15*(i%13))
      else
        row:SetPoint("TOPLEFT", FLMain, "TOPLEFT", 18, -53-15*13)
      end
      rowbuttons[i] = row
    end
  end
end


function ui:RosterPage(pg)
  rosterpage = rosterpage + pg
  for i, button in ipairs(rowbuttons) do
    if i <= #db.sortlist and i > rosterpage * 13 and i <= (rosterpage + 1) * 13 then
      button:Show()
    else
      button:Hide()
    end
  end
  if rosterpage == 0 then
    FLMainPrev:Hide()
  else
    FLMainPrev:Show()
  end
  if (rosterpage + 1) * 13 >= #db.sortlist then
    FLMainNext:Hide()
  else
    FLMainNext:Show()
  end
end

-------------------------------------------------------------------------------
--- ROSTER
-------------------------------------------------------------------------------
-- Handles UI for managing roster, not including standings-specific UI

function ui:InitializeRoster()
  FLRosterFrame:SetScale(.8)
  FLRosterFrame.TitleText:SetText("Remove player")
  CreateFrame("Button", "FLRosterDropDown", FLRosterFrame, "UIDropDownMenuTemplate")
  FLRosterDropDown:SetPoint("LEFT")
end


function ui:AddPlayerEntered(done)
  --- Parse & validate input, create player or print error
  local name = flsys:Cap(FLAddPlayerPlayer:GetText())
  local ep = FLAddPlayerEP:GetNumber()
  local gp = FLAddPlayerGP:GetNumber()
  if name=="" then
    flsys:Print("Please enter a player name.")
  elseif db.players[name] and not db.players[name].temp then
    flsys:Print(name.." is already in the system.  If you want to reset their data, remove them and then add.")
  else
    flsys:Debug("Adding player: "..name..", EP: "..ep..", GP: "..gp)
    flsys:NewPlayer(name, nil, nil, nil, ep, gp)
    FLAddPlayerPlayer:SetText("")
    FLAddPlayerEP:SetNumber(0)
    FLAddPlayerGP:SetNumber(0)
    if done then
      FLAddPlayer:Hide()
      flsys:RefreshSortList()
    end
  end
end

local function initialize_dropdown_all(self, level)
  -- used in raid for editing an award - shows all players even those not in db
  local button = {}
  for k,v in pairs(db.sortlist) do
    button.text = v
    button.func = function(self)
      UIDropDownMenu_SetSelectedName(FLRosterDropDown, self:GetText());
    end
    UIDropDownMenu_AddButton(button, level)
  end
end

local function initialize_dropdown_perm(self, level)
  -- shows only players that have been added to the system. used to remove players.
  local button = {}
  for k,v in pairs(db.sortlist) do
    if db.players[v] and not db.players[v].temp then
      button.text = v
      button.func = function(self)
        UIDropDownMenu_SetSelectedName(FLRosterDropDown, self:GetText());
      end
      UIDropDownMenu_AddButton(button, level)
    end
  end
end

function ui:InitializeAwardDropDown(current)
  if current then
    UIDropDownMenu_Initialize(FLRosterDropDown, initialize_dropdown_all)
  else
    UIDropDownMenu_Initialize(FLRosterDropDown, initialize_dropdown_perm)
  end
end

function ui:ShowRemovePlayer()
  if db.inprogress then
    StaticPopup_Show('NOTDURINGRAID')
  else
    InterfaceOptionsFrame:Hide()
    FLRosterDropDown:ClearAllPoints()
    FLRosterDropDown:SetParent(FLRosterFrame)
    FLRosterDropDown:SetPoint('CENTER')
    UIDropDownMenu_Initialize(FLRosterDropDown, initialize_dropdown_perm)
    FLRosterFrame:Show()
  end
end



-------------------------------------------------------------------------------
--- LOG
-------------------------------------------------------------------------------
-- Handles log frame, including popup export window

local log_rows = {} -- log row buttons

function ui:InitializeLog()
  FLLogBox:SetPoint('TOPRIGHT', FLLogContainer, 'TOPRIGHT', -4, -25)
  FLLogBoxScrollChildFrame:SetPoint('TOPRIGHT', FLLogBox, 'TOPRIGHT', -25, 0)
  FLLogBoxScrollChildFrame:SetPoint('BOTTOMLEFT', FLLogBox, 'BOTTOMLEFT', 0, 0)
  FLLogBox:SetScript('OnVerticalScroll', function(self, offset)
    ui:LogScrolled(self, offset, 18)
  end)
  FLLogItemAmount:SetNumeric(Enable)
  FLLogAdd:SetScale(.8)
  FLLogExport:SetScript('OnClick', function() ui:ShowExportFrame('gear') end)

  FLExport.TitleText:SetText('Copy and paste text to spreadsheet')
  FLExport.TitleText:SetJustifyH('LEFT')
end


function ui:ShowLogFrame()
  if #log_rows == 0 then --create frames if needed
    local fontstr
    for i = 1, 13 do
      log_rows[i] = CreateFrame('Button', 'LogRow'..i, FLLogContainer, 'LogRowTemplate')
      log_rows[i]:SetPoint('TOPLEFT', FLLogContainer, 'TOPLEFT', 5, -18*i + 15)
      log_rows[i]:SetMotionScriptsWhileDisabled(true)
      fontstr = log_rows[i]:GetFontString()
      fontstr:SetJustifyH('LEFT')
    end
    -- scroll down to newest content
    flsys:Debug('LOG DISPLAY', 'created rows, calling on vert scroll')
    local starting_os = math.max(#flsys.db.epgplog - 13, 0)
    ui:UpdateLogFrame(starting_os)
    FLLog:Show()
    FLLogBoxScrollBar:SetValue(starting_os*18)
  elseif not FLLog:IsShown() then
    ui:UpdateLogFrame()
    FLLog:Show()
  end
end

function ui:UpdateLogFrame(offset)
  --- given the position of the scroll, set each log row's text and onclick
  -- @offset: given in number of rows. calculated by OnVerticalScroll
  -- calls: FauxScrollFrame_Update
  -- called by: ShowLogFrame, OnVerticalScroll

  local num_logs = #flsys.db.epgplog
  local start_raid_i = 1 --for decay calculations. 1 = current
  if num_logs > 0 then
    for i = 1, num_logs - (offset or 0) do
      if flsys.db.epgplog[i][3] == 'decay' then
        start_raid_i = start_raid_i + 1
      end
    end
  end
  for i = 1, 13 do
    local log_i = num_logs - (i-1) - (offset or 0)
    local raid_i = start_raid_i -- so on script don't all call the last version
    if log_i > 0 then
      local logitem = flsys.db.epgplog[log_i]
      log_rows[i]:SetText(flsys:LogItemtoString(logitem))
      fs = log_rows[i]:GetFontString()
      if fs:IsTruncated() then
        log_rows[i]:SetScript('OnEnter', function(self)
          GameTooltip:SetOwner(self, 'ANCHOR_RIGHT')
          GameTooltip:SetText(self:GetText())
          GameTooltip:Show()
        end)
        log_rows[i]:SetScript('OnLeave', function() GameTooltip:Hide() end)
      else
        log_rows[i]:SetScript('OnEnter', nil)
        log_rows[i]:SetScript('OnLeave', nil)
      end
      if logitem[3] == 'GP' or (logitem[3] == 'EP' and not string.find(logitem[4], ',')) then
        -- can edit/delete
        --flsys:Print(i..'gp or single ep, setting texture')
        log_rows[i]:Enable()
        log_rows[i]:SetNormalTexture('Interface\\Buttons\\UI-GuildButton-PublicNote-Up')
        log_rows[i]:SetPushedTexture('Interface\\Buttons\\UI-GuildButton-PublicNote-Up')
        log_rows[i]:SetScript('OnClick', function() ui:ShowEditLogItem(log_i, raid_i); end)
      elseif logitem[3] == 'start' then
        -- can add
        log_rows[i]:Enable()
        log_rows[i]:SetNormalTexture('Interface\\Buttons\\UI-AttributeButton-Encourage-Up')
        log_rows[i]:SetPushedTexture('Interface\\Buttons\\UI-AttributeButton-Encourage-Down')
        log_rows[i]:SetScript('OnClick', function() ui:ShowAddType(log_i, raid_i); end)
      else
        log_rows[i]:Disable()
        if logitem[3] == 'decay' then
          start_raid_i = start_raid_i - 1
        end
      end
      if not log_rows[i]:IsShown() then
        log_rows[i]:Show()
      end
    else
      -- extra row doesn't correspond to any log entry; hide it
      log_rows[i]:Hide()
    end
  end
  FauxScrollFrame_Update(FLLogBox, num_logs, 13, 18)
end

function ui:LogScrolled(self, value, itemHeight)
	FLLogBoxScrollBar:SetValue(value)
  offset = floor((value / itemHeight) + 0.5);
	self.offset = offset
  ui:UpdateLogFrame(offset)
end

function ui:LogChanged()

  -- update log frame if needed
  if FLLog:IsShown() then
    ui:UpdateLogFrame()
  end

  -- update log edit/add if needed
  if FLLogItem:IsShown() then
    local raid_i = FLLogItem:GetAttribute("raid")
    local log_i = FLLogItem:GetAttribute("log") + 1
    if logtype=="decay" then
      raid_i = raid_i + 1
    end
    ui:ShowEditLogItem(log_i, raid_i)
  end

end

function ui:ShowAddType(log_i, raid_i)
  FLLogAddEP:SetScript('OnClick', function() ui:ShowAddToLog(log_i, raid_i, 'EP') end)
  FLLogAddGP:SetScript('OnClick', function() ui:ShowAddToLog(log_i, raid_i, 'GP') end)
  --FLLogAddMassEP:SetScript('OnClick', function() ui:ShowAddToLog(log_i, raid_i, 'EPmass') end)

  FLLogAdd:Show()
end

local function setup_dropdown(raid_i)
  -- used when adding to log or editing log item

  FLRosterDropDown:ClearAllPoints()
  FLRosterDropDown:SetParent(FLLogItem)
  FLRosterDropDown:SetPoint('BOTTOMLEFT', FLLogItem, 'BOTTOMLEFT', 70, 30)

  ui:InitializeAwardDropDown(raid_i == 1)

end


function ui:ShowAddToLog(log_i, raid_i, award_type)
  --- Shows FLLogItem, set up for adding an award retroactively
  -- @log_i: the index of the epgplog item where the raid started
  -- @raid_i: raid index, used to scale and place the award appropriately
  -- @award_type: ep or gp

  flsys:Info('Showing FLLogItem for adding something retroactively. Type: '..award_type)
  FLLogAdd:Hide()
  local start = flsys.db.epgplog[log_i]

  FLLogItem.TitleText:SetText('Adding award to '..date('%m/%d/%y', start[1])
                               ..' '..string.sub(start[5], 1, -8))
  FLLogItemAmount:SetNumber(0)
  FLLogItemType:SetText(award_type)
  FLLogItemReason:SetText('')
  FLLogItem:SetHeight(105)
  FLLogItemNote:SetText('')

  setup_dropdown(raid_i)

  FLLogItemDelete:Hide()
  FLLogItemUpdate:SetScript('OnClick', function() flsys:AlterLog(log_i, raid_i, 'add') end)
  FLLogItem:Show()
end


function ui:ShowEditLogItem(log_i, raid_i)
  --- Shows FLLogItem, set up for editing/deleting a log entry

  local logitem = flsys.db.epgplog[log_i]

  local desc = 'Editing award '..date('%m/%d/%y %H:%M', logitem[1])
  flsys:Info(desc)
  FLLogItem.TitleText:SetText(desc)
  FLLogItemAmount:SetNumber(logitem[2])
  FLLogItemType:SetText(logitem[3])
  FLLogItemReason:SetText(logitem[5])
  FLLogItem:SetAttribute('raid', raid_i) --TODO TEST
  FLLogItem:SetAttribute('log', log_i)

  setup_dropdown(raid_i)
  if flsys.db.players[logitem[4]] and (raid_i == 1 or not flsys.db.players[logitem[4]].temp) then
    FLLogItem:SetHeight(105)
    FLLogItemNote:SetText('')
    UIDropDownMenu_SetSelectedName(FLRosterDropDown, logitem[4])
  else
    FLLogItem:SetHeight(145)
    if flsys.db.alts_data[logitem[4]] and logtype[3] == 'GP' then
      -- EP awards for alts are stored under main's name, but GP are not
      FLLogItemNote:SetText('Note: This was a GP award to an alt, which has no impact on ongoing PR.')
    else
      FLLogItemNote:SetText('Note: This player is no longer in the database. Editing the amount or reason won\'t impact anyone\'s PR, but you can reassign the award to another player.')
    end
  end

  FLLogItemDelete:SetScript('OnClick', function()
    local dialog = StaticPopup_Show('DELETE_LOG_ITEM')
    if dialog then
      dialog.log_i = log_i
      dialog.raid_i = raid_i
    end
  end)
  FLLogItemDelete:Show()
  FLLogItemUpdate:SetScript('OnClick', function() flsys:AlterLog(log_i, raid_i, 'edit') end)
  FLLogItem:Show()
end



-- TODO can we guarantee sortlist is up to date?
function ui:ShowExportFrame(exporttype)
  -- epgplog format: 1 = timestamp, 2 = amount, 3 = type,
  --                 4 = player(s), 5 = reason/item, 6 = itemid
  -- spreadsheet order: 1 4 5 2 3 (6)
  local exporttext = ''

  if exporttype == 'gear' then
    for i, entry in ipairs(flsys.db.epgplog) do
      exporttext = exporttext..date('%m/%d/%y %H:%M', entry[1])..'\t'..entry[4]..'\t'..entry[5]..'\t'..entry[2]..'\t'..entry[3]
      if entry[6] then
        exporttext = exporttext..'\t'..entry[6]
      end
      exporttext = exporttext..'\n'
    end
    FLExportDone:SetScript('OnClick', function() FLExport:Hide(); flsys:ConfirmExportTime('epgp') end)

  elseif exporttype == 'standings' then
    local ddate = date('%m/%d/%y %H:%M', time())
    for i, name in ipairs(flsys.db.sortlist) do
      exporttext = exporttext..name..'\t'..flsys.db.players[name].ep_total..'\t'..flsys.db.players[name].gp_total..'\t'..flsys.db.players[name].current_pr..'\t'..ddate
      if i == 1 then
        exporttext = exporttext..'\t'..flsys:ExportToString()
      end
      exporttext = exporttext..'\n'
    end
    FLExportDone:SetScript('OnClick', function() FLExport:Hide(); end)
  end

  if exporttext ~= '' then
    FLExportText:SetText(exporttext)
    FLExport:Show()
  else
    self:Print('Nothing to export.')
  end
end

function ui:ConfirmExportTime(exporttype)
  -- Asks the player to confirm export complete so we can trim log
  dialog = StaticPopup_Show('EXPORT_COMPLETE', exporttype)
  if dialog then
    dialog.etype = exporttype
  end
end

-------------------------------------------------------------------------------
--- OPTIONS
-------------------------------------------------------------------------------

function ui:InitializeOptions()
  -- import/add
  FLImportData.TitleText:SetText('Import data')
  FLAddPlayer.TitleText:SetText('Add player')
  FLAddPlayer:SetScale(0.8)
  FLAddPlayerEP:SetNumeric(enable)
  FLAddPlayerGP:SetNumeric(enable)
  FLAddPlayerEP:SetNumber(0)
  FLAddPlayerGP:SetNumber(0)
  FLAddPlayerGP:SetScript("OnEnterPressed", function()
    flsys.ui:AddPlayerEntered(true)
  end)
  FLAddPlayerMore:SetScript("OnClick", function()
    flsys.ui:AddPlayerEntered(true)
  end)
  FLAddPlayerDone:SetScript("OnClick", function()
    flsys.ui:AddPlayerEntered(false)
  end)
  FLImportData:SetScale(.8)
end

function ui:ShowAddPlayer()
  InterfaceOptionsFrame:Hide()
  FLAddPlayer:Show()
end

function ui:ShowImport()
  InterfaceOptionsFrame:Hide()
  FLImportDataData:SetText('')
  FLImportData:Show()
end

